import datetime
import time

import pytest

from hiro.errors import InvalidTypeError
from hiro.utils import chained, time_in_seconds, timedelta_to_seconds, utc


def test_fractional():
    delta = datetime.timedelta(seconds=1, microseconds=1000)
    assert round(abs(timedelta_to_seconds(delta) - 1.001), 7) == 0


def test_days():
    delta = datetime.timedelta(days=10)
    assert timedelta_to_seconds(delta) == delta.days * 24 * 60 * 60


def test_passthrough():
    assert time_in_seconds(1) == 1

    long_instance = 1
    assert time_in_seconds(long_instance) == long_instance
    assert time_in_seconds(1.0) == 1.0


def test_date():
    d = datetime.date(1970, 1, 1)
    assert time_in_seconds(d) == 0

    d = d + datetime.timedelta(days=1234)
    assert time_in_seconds(d) == 1234 * 24 * 60 * 60


def test_datetime():
    d = datetime.datetime(1970, 1, 1, 0, 0, 0)
    assert time_in_seconds(d) == time.mktime(d.timetuple())


def test_tzaware_datetime():
    d = datetime.datetime(1970, 1, 1, 0, 0, 0, tzinfo=utc)
    assert time_in_seconds(d) == 0


def test_invalid_type():
    with pytest.raises(InvalidTypeError):
        time_in_seconds("this is a string")


class TestChained:
    class Foo:
        @chained
        def return_value(self, value=None):
            return value

    def setup_method(self):
        self.obj = self.Foo()

    def test_no_return(self):
        assert self.obj.return_value() is self.obj

    def test_with_return(self):
        o = object()
        assert self.obj.return_value(o) is o

    def test_kwargs(self):
        o = object()
        assert self.obj.return_value(value=o) is o



"""
timeline & runner implementation
"""
import copy
import datetime
import inspect
import sys
import threading
import time
from functools import wraps
from unittest import mock

from .errors import SegmentNotComplete, TimeOutofBounds
from .patches import Date, Datetime
from .utils import chained, time_in_seconds, timedelta_to_seconds

IGNORED_MODULES = set()
_NO_EXCEPTION = (None, None, None)


class Decorator:
    def __call__(self, fn):
        @wraps(fn)
        def inner(*args, **kw):
            self.__enter__()
            exc = _NO_EXCEPTION
            try:
                if "timeline" in inspect.signature(fn).parameters:
                    result = fn(*args, timeline=self, **kw)
                else:
                    result = fn(*args, **kw)
            except Exception:
                exc = sys.exc_info()

            catch = self.__exit__(*exc)

            if not catch and exc is not _NO_EXCEPTION:
                raise exc[0]

            return result

        return inner


class Segment:
    """
    utility class to manage execution result and timings
    for :class:`SyncRunner
    """

    def __init__(self):
        self.__error = False
        self.__response = None
        self.__start = time.time()
        self.__end = None

    def complete(self, response):
        """
        called upon successful completion of the segment
        """
        self.__response = response

    def complete_with_error(self, exception):
        """
        called if the segment errored during execution
        """
        self.__error = exception

    @property
    def complete_time(self):
        """
        returns the completion time
        """

        return self.__end

    @complete_time.setter
    def complete_time(self, completion_time):
        """
        sets the completion time
        """
        self.__end = completion_time

    @property
    def start_time(self):
        """
        returns the start time
        """

        return self.__start

    @start_time.setter
    def start_time(self, start_time):
        """
        sets the start time
        """
        self.__start = start_time

    @property
    def runtime(self):
        """
        returns the total execution time of the segment
        """

        if self.__end:
            return self.complete_time - self.start_time
        else:
            raise SegmentNotComplete

    @property
    def response(self):
        """
        returns the return value captured in the segment or raises
        the :exception:`exceptions.Exception` that was caught.
        """

        if not self.__end:
            raise SegmentNotComplete
        else:
            if self.__error:
                raise self.__error[0]

            return self.__response


class Timeline(Decorator):
    """
    Timeline context manager. Within this context
    the following builtins respect the alterations made
    to the timeline:

    - :func:`time.time`
    - :func:`time.time_ns`
    - :func:`time.monotonic`
    - :func:`time.monotonic_ns`
    - :func:`time.sleep`
    - :func:`time.localtime`
    - :func:`time.gmtime`
    - :meth:`datetime.datetime.now`
    - :meth:`datetime.date.today`
    - :meth:`datetime.datetime.utcnow`

    The class can be used either as a context manager or a decorator.

    The following are all valid ways to use it.

    .. code-block:: python

        with Timeline(scale=10, start=datetime.datetime(2012,12,12)):
            ....

        fast_timeline = Timeline(scale=10).forward(120)

        with fast_timeline as timeline:
            ....

        delta = datetime.date(2015,1,1) - datetime.date.today()
        future_frozen_timeline = Timeline(scale=10000).freeze().forward(delta)
        with future_frozen_timeline as timeline:
            ...

        @Timeline(scale=100)
        def slow():
            time.sleep(120)


    :param float scale: > 1 time will go faster and < 1 it will be slowed down.
    :param start: if specified starts the timeline at the given value (either a
        floating point representing seconds since epoch or a
        :class:`datetime.datetime` object)

    """

    class_mappings = {
        "date": (datetime.date, Date),
        "datetime": (datetime.datetime, Datetime),
    }

    def __init__(self, scale=1, start=None):
        self.reference = time.time()
        self.offset = (
            time_in_seconds(start) - self.reference if start is not None else 0.0
        )
        self.freeze_point = self.freeze_at = None
        self.patchers = []
        self.mock_mappings = {
            "datetime.date": (datetime.date, Date),
            "datetime.datetime": (datetime.datetime, Datetime),
            "time.monotonic": (time.monotonic, self.__time_monotonic),
            "time.monotonic_ns": (time.monotonic_ns, self.__time_monotonic_ns),
            "time.time": (time.time, self.__time_time),
            "time.time_ns": (time.time_ns, self.__time_time_ns),
            "time.sleep": (time.sleep, self.__time_sleep),
            "time.gmtime": (time.gmtime, self.__time_gmtime),
            "time.localtime": (time.localtime, self.__time_localtime),
        }
        self.func_mappings = {
            "time": (time.time, self.__time_time),
            "time_ns": (time.time_ns, self.__time_time_ns),
            "monotonic": (time.monotonic, self.__time_monotonic),
            "monotonic_ns": (time.monotonic_ns, self.__time_monotonic_ns),
            "sleep": (time.sleep, self.__time_sleep),
            "gmtime": (time.gmtime, self.__time_gmtime),
            "localtime": (time.localtime, self.__time_localtime),
        }
        self.factor = scale

    def _get_original(self, fn_or_mod):
        """
        returns the original moduel or function
        """

        if fn_or_mod in self.mock_mappings:
            return self.mock_mappings[fn_or_mod][0]
        elif fn_or_mod in self.func_mappings:
            return self.func_mappings[fn_or_mod][0]
        else:
            return self.class_mappings[fn_or_mod][0]

    def _get_fake(self, fn_or_mod):
        """
        returns the mocked/patched module or function
        """

        if fn_or_mod in self.mock_mappings:
            return self.mock_mappings[fn_or_mod][1]
        elif fn_or_mod in self.func_mappings:
            return self.func_mappings[fn_or_mod][1]
        else:
            return self.class_mappings[fn_or_mod][1]

    def __compute_time(self, freeze_point, offset, original, unit=1, cast_func=float):
        """
        computes the current_time after accounting for
        any adjustments due to :attr:`factor` or invocations
        of :meth:`freeze`, :meth:`rewind` or :meth:`forward`
        """
        if freeze_point is not None:
            return unit * (offset + freeze_point)
        else:
            delta = self._get_original(original)() - (unit * self.reference)
            return cast_func(
                unit * self.reference + (delta * self.factor) + unit * offset
            )

    def __check_out_of_bounds(self, offset=None, freeze_point=None):
        """
        ensures that the time that would be calculated based on any
        offset or freeze point would not result in jumping beyond the epoch
        """
        next_time = self.__compute_time(
            freeze_point or self.freeze_point, offset or self.offset, "time.time"
        )

        if next_time < 0:
            raise TimeOutofBounds(next_time)

    def __time_monotonic(self):
        """
        patched version of :func:`time.monotonic`
        """

        return self.__compute_time(self.freeze_point, self.offset, "time.monotonic")

    def __time_monotonic_ns(self):
        """
        patched version of :func:`time.monotonic_ns`
        """

        return self.__compute_time(
            self.freeze_point, self.offset, "time.monotonic_ns", 1e9, int
        )

    def __time_time(self):
        """
        patched version of :func:`time.time`
        """

        return self.__compute_time(self.freeze_point, self.offset, "time.time")

    def __time_time_ns(self):
        """
        patched version of :func:`time.time_ns`
        """

        return self.__compute_time(
            self.freeze_point, self.offset, "time.time_ns", 1e9, int
        )

    def __time_gmtime(self, seconds=None):
        """
        patched version of :func:`time.gmtime`
        """

        return self._get_original("time.gmtime")(
            seconds if seconds is not None else self.__time_time()
        )

    def __time_localtime(self, seconds=None):
        """
        patched version of :func:`time.localtime`
        """

        return self._get_original("time.localtime")(
            seconds if seconds is not None else self.__time_time()
        )

    def __time_sleep(self, amount):
        """
        patched version of :func:`time.sleep`
        """
        self._get_original("time.sleep")(1.0 * amount / self.factor)

    @chained
    def forward(self, amount):
        """
        forwards the timeline by the specified :attr:`amount`

        :param amount: either an integer representing seconds or
         a :class:`datetime.timedelta` object
        """
        offset = self.offset

        if isinstance(amount, datetime.timedelta):
            offset += timedelta_to_seconds(amount)
        else:
            offset += amount
        self.__check_out_of_bounds(offset=offset)
        self.offset = offset

    @chained
    def rewind(self, amount):
        """
        rewinds the timeline by the specified :attr:`amount`

        :param amount: either an integer representing seconds or
         a :class:`datetime.timedelta` object
        """
        offset = self.offset

        if isinstance(amount, datetime.timedelta):
            offset -= timedelta_to_seconds(amount)
        else:
            offset -= amount
        self.__check_out_of_bounds(offset=offset)
        self.offset = offset

    @chained
    def freeze(self, target_time=None):
        """
        freezes the timeline

        :param target_time: the time to freeze at as either a float
          representing seconds since the epoch or a :class:`datetime.datetime`
          object. If not provided time will be frozen at the current time of
          the enclosing :class:`Timeline`
        """

        if target_time is None:
            freeze_point = self._get_fake("time.time")()
        else:
            freeze_point = time_in_seconds(target_time)
        self.__check_out_of_bounds(freeze_point=freeze_point)
        self.freeze_point = freeze_point
        self.offset = 0

    @chained
    def unfreeze(self):
        """
        if a call to :meth:`freeze` was previously made, the timeline will be
        unfrozen to the point which :meth:`freeze` was invoked.

        .. warning::

            Since unfreezing will reset the timeline back to the point in
            when the :meth:`freeze` was invoked - the effect of previous
            invocations of :meth:`forward` and :meth:`rewind` will
            be lost. This is by design so that freeze/unfreeze can be used as
            a checkpoint mechanism.

        """

        if self.freeze_point is not None:
            self.reference = self._get_original("time.time")()
            self.offset = time_in_seconds(self.freeze_point) - self.reference
            self.freeze_point = None

    @chained
    def scale(self, factor):
        """
        changes the speed at which time elapses and how long sleeps last for.

        :param float factor: > 1 time will go faster and < 1 it will be slowed
            down.

        """
        self.factor = factor
        self.reference = self._get_original("time.time")()

    @chained
    def reset(self):
        """
        resets the current timeline to the actual time now
        with a scale factor 1
        """

        self.factor = 1
        self.freeze_point = None
        self.reference = self._get_original("time.time")()
        self.offset = 0

    def __enter__(self):
        for name in list(sys.modules.keys()):
            module = sys.modules[name]

            if module in IGNORED_MODULES:
                continue
            mappings = copy.copy(self.class_mappings)
            mappings.update(self.func_mappings)

            try:
                for obj in mappings:
                    if obj in dir(module) and getattr(
                        module, obj
                    ) == self._get_original(obj):
                        path = "{}.{}".format(name, obj)

                        if path not in self.mock_mappings:
                            patcher = mock.patch(path, self._get_fake(obj))
                            patcher.start()
                            self.patchers.append(patcher)
            # this is done for cases where invalid modules are on
            # sys modules.
            except:  # noqa: E722
                IGNORED_MODULES.add(module)

        for time_obj in self.mock_mappings:
            patcher = mock.patch(time_obj, self._get_fake(time_obj))
            patcher.start()
            self.patchers.append(patcher)

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        for patcher in self.patchers:
            patcher.stop()
        self.patchers = []


class ScaledRunner:
    """
    manages the execution of a callable within a :class:`hiro.Timeline`
    context.
    """

    def __init__(self, factor, func, *args, **kwargs):
        self.func = func
        self.func_args = args
        self.func_kwargs = kwargs
        self.segment = Segment()
        self.factor = factor
        self.__call__()

    def _run(self):
        """
        managed execution of :attr:`func`
        """
        self.segment.start_time = time.time()
        with Timeline(scale=self.factor):
            try:
                self.segment.complete(self.func(*self.func_args, **self.func_kwargs))
            # will be rethrown
            except:  # noqa: E722
                self.segment.complete_with_error(sys.exc_info())
        self.segment.complete_time = time.time()

    def __call__(self):
        self._run()

        return self

    def get_response(self):
        """
        :returns: the return value from :attr:`func`
        :raises: Exception if the :attr:`func` raised one during execution
        """

        return self.segment.response

    def get_execution_time(self):
        """
        :returns: the real execution time of :attr:`func` in seconds
        """

        return self.segment.runtime


class ScaledThreadedRunner(ScaledRunner):
    """
    manages the threaded execution of a callable within a
    :class:`hiro.Timeline` context.
    """

    def __init__(self, *args, **kwargs):
        self.thread_runner = threading.Thread(target=self._run)
        super().__init__(*args, **kwargs)

    def __call__(self):
        self.thread_runner.start()

        return self

    def is_running(self):
        """
        :rtype bool: whether the :attr:`func` is still running or not.
        """

        return self.thread_runner.is_alive()

    def join(self):
        """
        waits for the :attr:`func` to complete execution.
        """

        return self.thread_runner.join()


def run_sync(factor, func, *args, **kwargs):
    """
    Executes a callable within a :class:`hiro.Timeline`

    :param int factor: scale factor to use for the timeline during execution
    :param function func: the function to invoke
    :param args: the arguments to pass to the function
    :param kwargs: the keyword arguments to pass to the function
    :returns: an instance of :class:`hiro.core.ScaledRunner`

    """

    return ScaledRunner(factor, func, *args, **kwargs)


def run_threaded(factor, func, *args, **kwargs):
    """
    Executes a callable in a separate thread within a :class:`hiro.Timeline`

    :param int factor: scale factor to use for the timeline during execution
    :param function func: the function to invoke
    :param args: the arguments to pass to the function
    :param kwargs: the keyword arguments to pass to the function
    :returns: an instance of :class:`hiro.core.ScaledThreadedRunner`

    """

    return ScaledThreadedRunner(factor, func, *args, **kwargs)


# For backward compatibility


def run_async(factor, func, *args, **kwargs):
    """
    Executes a callable in a separate thread within a :class:`hiro.Timeline`

    :param int factor: scale factor to use for the timeline during execution
    :param function func: the function to invoke
    :param args: the arguments to pass to the function
    :param kwargs: the keyword arguments to pass to the function
    :returns: an instance of :class:`hiro.core.ScaledAsyncRunner`

    .. deprecated:: 1.0.0
       Use :meth:`run_threaded`

    """

    return run_threaded(factor, func, *args, **kwargs)


ScaledAsyncRunner = ScaledThreadedRunner



import math
import os
import time
from datetime import date, datetime, timedelta
from unittest import mock

import pytest

from hiro import Timeline
from hiro.utils import timedelta_to_seconds
from tests.emulated_modules import sample_1, sample_2, sample_3


def test_accelerate():
    s = time.time()
    with Timeline(100):
        time.sleep(10)
    assert time.time() - s < 10


def test_deccelerate():
    s = time.time()
    with Timeline(0.5):
        time.sleep(0.25)
    assert time.time() - s > 0.25


def test_check_time():
    start = time.time()
    with Timeline(100):
        for i in range(0, 10):
            time.sleep(1)
            delta = time.time() - (start + (i + 1) * 100)
            assert delta <= 1.5, delta


def test_utc():
    start_local = datetime.now()
    start_utc = datetime.utcnow()
    with Timeline(100000):
        time.sleep(60 * 60)
        future_diff = (datetime.now() - datetime.utcnow()).seconds / 60.0 / 60.0
        start_diff = (start_local - start_utc).seconds / 60.0 / 60.0
        assert math.ceil(future_diff) == math.ceil(start_diff)

        time.sleep(60 * 60 * 23)
        assert (datetime.now() - start_local).days == 1
        assert (datetime.utcnow() - start_utc).days == 1


def test_forward():
    original = date.today()
    with Timeline() as timeline:
        timeline.forward(86400)
        assert (date.today() - original).days == 1


def test_rewind():
    original_day = date.today()
    original_datetime = datetime.now()
    with Timeline() as timeline:
        timeline.rewind(86400)
        assert (original_day - date.today()).days == 1
        timeline.forward(timedelta(days=1))
        assert (original_day - date.today()).days == 0
        timeline.rewind(timedelta(days=1))
        amount = original_datetime - datetime.now()
        amount_seconds = timedelta_to_seconds(amount)
        assert int(amount_seconds) >= 86399, amount_seconds


def test_freeze():
    with Timeline() as timeline:
        timeline.freeze()
        originals = (
            sample_1.sub_module_1.sub_sample_1_1_time(),
            sample_1.sub_module_1.sub_sample_1_1_now(),
        )
        time.sleep(1)
        assert time.time() == originals[0]
        assert datetime.now() == originals[1]
    with Timeline() as timeline:
        timeline.freeze()
        originals = (
            sample_2.sub_module_2.sub_sample_2_1_time(),
            sample_2.sub_module_2.sub_sample_2_1_now(),
        )
        time.sleep(1)
        assert time.time() == originals[0]
        assert datetime.now() == originals[1]
    with Timeline() as timeline:
        timeline.freeze()
        originals = (
            sample_3.sub_module_3.sub_sample_3_1_time(),
            sample_3.sub_module_3.sub_sample_3_1_now(),
            sample_3.sub_module_3.sub_sample_3_1_gmtime(),
        )
        time.sleep(1)
        assert time.time() == originals[0]
        assert datetime.now() == originals[1]
        assert time.gmtime() == originals[2]


def test_freeze_target():
    with Timeline() as timeline:
        timeline.freeze(datetime.fromtimestamp(0))
        assert round(abs(time.time() - 0), 1) == 0

    with Timeline() as timeline:
        timeline.freeze(0)
        assert round(abs(time.time() - 0), 1) == 0
        assert round(abs(time.time_ns() - 0), 1) == 0
        assert round(abs(time.monotonic() - 0), 1) == 0
        assert round(abs(time.monotonic_ns() - 0), 1) == 0

    with Timeline() as timeline:
        with pytest.raises(TypeError):
            timeline.freeze("now")

    with Timeline() as timeline:
        with pytest.raises(AttributeError):
            timeline.freeze(-1)
        with pytest.raises(AttributeError):
            timeline.freeze(date(1969, 1, 1))


def test_unfreeze():
    real_day = date.today()
    with Timeline() as timeline:
        timeline.freeze(0)
        assert round(abs(time.time() - 0), 7) == 0
        timeline.unfreeze()
        timeline.scale(10)
        time.sleep(1)
        assert int(time.time()) == 1
        timeline.forward(timedelta(minutes=2))
        assert int(time.time()) == 121
        timeline.reset()
        assert (int(time.time()) - int(os.popen("date +%s").read().strip())) <= 1
        timeline.forward(timedelta(days=1))
        assert (date.today() - real_day).days == 1


def test_freeze_forward_unfreeze():

    # start at 2012/12/12 0:0:0
    test_timeline = Timeline(scale=100, start=datetime(2012, 12, 12, 0, 0, 0))

    # jump forward an hour and freeze
    with test_timeline.forward(60 * 60).freeze():
        assert (
            round(
                abs((datetime.now() - datetime(2012, 12, 12, 1, 0, 0)).seconds - 0), 7
            )
            == 0
        )
        # sleep 10 seconds
        time.sleep(10)
        # assert no changes
        assert (
            round(
                abs((datetime.now() - datetime(2012, 12, 12, 1, 0, 0)).seconds - 0), 7
            )
            == 0
        )
        # unfreeze timeline
        test_timeline.unfreeze()
        # ensure unfreeze was at the original freeze point
        assert (
            round(
                abs((datetime.now() - datetime(2012, 12, 12, 1, 0, 0)).seconds - 0), 7
            )
            == 0
        )
        # sleep 10 seconds
        time.sleep(10)
        # ensure post unfreeze, time moves
        assert (datetime.now() - datetime(2012, 12, 12, 1, 0, 0)).seconds >= 10

        # ensure post unfreeze, forward operations work
        test_timeline.forward(timedelta(hours=2))
        assert (
            int((datetime.now() - datetime(2012, 12, 12, 1, 0, 0)).seconds / 3600) == 2
        )
        # reset everything
        test_timeline.reset()
        assert int(time.time()) == int(os.popen("date +%s").read().strip())


def test_fluent():
    start = datetime.now()
    with Timeline().scale(10).forward(120):
        assert int(timedelta_to_seconds(datetime.now() - start)) >= 120
        time.sleep(10)
        assert int(timedelta_to_seconds(datetime.now() - start)) >= 130
    assert (datetime.now() - start).seconds < 10


def test_decorated():
    start = datetime(2013, 1, 1, 0, 0, 0)
    real_start = datetime.now()

    @Timeline(scale=10, start=start)
    def _decorated():
        time.sleep(10)
        assert int(timedelta_to_seconds(datetime.now() - start)) >= 10

    _decorated()
    assert (datetime.now() - real_start).seconds < 10


def test_decorated_with_argument():
    @Timeline()
    def _decorated(timeline):
        assert isinstance(timeline, Timeline)

    _decorated()

    @Timeline()
    def _decorated(timeline=None):
        assert isinstance(timeline, Timeline)

    _decorated()


def test_decorated_exception():
    class SomeException(Exception):
        pass

    @Timeline()
    def _decorated():
        raise SomeException("asdf")

    with pytest.raises(SomeException):
        _decorated()


@mock.patch("hiro.core.IGNORED_MODULES", new_callable=set)
def test_patch_ignored_modules(IGNORED_MODULES):
    hiro_dummy_module = mock.MagicMock(__dir__=mock.MagicMock(side_effect=Exception))

    with mock.patch.dict("sys.modules", {"hiro_dummy_module": hiro_dummy_module}):
        with Timeline().freeze():
            pass

        with Timeline().freeze():
            pass

        hiro_dummy_module.__dir__.assert_called_once()
        assert hiro_dummy_module in IGNORED_MODULES



"""

"""
from datetime import date, datetime
from time import gmtime, sleep, time


def sub_sample_3_1_now():
    return datetime.now()


def sub_sample_3_1_today():
    return date.today()


def sub_sample_3_1_sleep(seconds):
    sleep(seconds)


def sub_sample_3_1_time():
    return time()


def sub_sample_3_1_gmtime():
    return gmtime()



from .sub_sample_3_1 import (
    sub_sample_3_1_gmtime,
    sub_sample_3_1_now,
    sub_sample_3_1_sleep,
    sub_sample_3_1_time,
    sub_sample_3_1_today,
)

__all__ = [
    "sub_sample_3_1_now",
    "sub_sample_3_1_today",
    "sub_sample_3_1_sleep",
    "sub_sample_3_1_time",
    "sub_sample_3_1_gmtime",
]



"""

"""
import time

import pytest

import hiro
from hiro.errors import SegmentNotComplete


def test_scale_up_sync_runner():
    def _slow_func():
        time.sleep(1)
        return 1

    f = hiro.run_sync(4, _slow_func)
    assert f.get_execution_time() < 1
    assert f.get_response() == 1


def test_scale_up_sync_runner_fail():
    def _slow_func():
        time.sleep(1)
        raise Exception("foo")

    f = hiro.run_sync(4, _slow_func)
    with pytest.raises(Exception):
        f.get_response()
    assert f.get_execution_time() < 1


def test_scale_up_async_runner():
    def _slow_func():
        time.sleep(1)

    f = hiro.run_threaded(4, _slow_func)
    assert f.is_running()
    f.join()
    assert f.get_execution_time() < 1


def test_scale_up_async_runner_fail():
    def _slow_func():
        time.sleep(1)
        raise Exception("foo")

    f = hiro.run_threaded(4, _slow_func)
    assert f.is_running()
    f.join()
    with pytest.raises(Exception):
        f.get_response()
    assert f.get_execution_time() < 1


def test_segment_not_complete_error():
    def _slow_func():
        time.sleep(1)
        raise Exception("foo")

    f = hiro.run_threaded(4, _slow_func)
    with pytest.raises(SegmentNotComplete):
        f.get_execution_time()
    with pytest.raises(SegmentNotComplete):
        f.get_response()
    assert f.is_running()
    f.join()
    with pytest.raises(Exception):
        f.get_response()
    assert f.get_execution_time() < 1



"""
exceptions used by hiro.
"""


class SegmentNotComplete(Exception):
    """
    used to raise an exception if an async segment hasn't completed yet
    """


class TimeOutofBounds(AttributeError):
    """
    used to raise an exception when time is rewound beyond the epoch
    """

    def __init__(self, oob_time):
        message = (
            "you've frozen time at a point before the epoch (%d)."
            "hiro only supports going back to 1970/01/01 07:30:00" % oob_time
        )
        super().__init__(message)


class InvalidTypeError(TypeError):
    """
    used to raise an exception when an invalid type is provided
    for type operations
    """

    def __init__(self, value):
        message = (
            "%s provided when only float, int, datetime, or date objects"
            "are supported" % type(value)
        )
        super().__init__(message)



"""
random utility functions
"""
import calendar
import datetime
import functools
import time

from .errors import InvalidTypeError

utc = datetime.timezone.utc


def timedelta_to_seconds(delta):
    """
    converts a timedelta object to seconds
    """
    seconds = delta.microseconds
    seconds += (delta.seconds + delta.days * 24 * 3600) * 10**6
    return float(seconds) / 10**6


def time_in_seconds(value):
    """
    normalized either a datetime.date, datetime.datetime or float
    to a float corresponding to a UTC timestamp.

    datetime.date objects are converted to the timestamp of midnight UTC on
    that date.

    Naive datetime objects (without timezone information) are taken in the
    local timezone.
    """
    if isinstance(value, (float, int)):
        return value
    elif isinstance(value, datetime.datetime):
        if value.tzinfo is None:
            return time.mktime(value.timetuple())
        else:
            return calendar.timegm(value.astimezone(utc).timetuple())
    elif isinstance(value, datetime.date):
        # We're really explicit here but we could just use value.timetuple()
        return calendar.timegm(
            datetime.datetime.combine(value, datetime.time(0, 0, 0, 0, utc)).timetuple()
        )
    else:
        raise InvalidTypeError(value)


def chained(method):
    """
    Method decorator to allow chaining.

    adopted from:
    http://www.snip2code.com/Snippet/2535/Fluent-interface-decorators
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        """
        fluent wrapper
        """
        result = method(self, *args, **kwargs)
        return self if result is None else result

    return wrapper



"""
time manipulation utilities for python
"""

from . import _version
from .core import Timeline, run_async, run_sync, run_threaded

__all__ = ["run_threaded", "run_async", "run_sync", "Timeline"]

__version__ = _version.get_versions()["version"]



"""

"""
import datetime
import time


def sub_sample_1_1_now():
    return datetime.datetime.now()


def sub_sample_1_1_today():
    return datetime.date.today()


def sub_sample_1_1_sleep(seconds):
    datetime.time.sleep(seconds)


def sub_sample_1_1_time():
    return time.time()



"""
patched builtin time classes for use by :class:`hiro.Timeline`
"""
import abc
import time
from datetime import date as realdate
from datetime import datetime as realdatetime


class DatetimeMeta(abc.ABCMeta):
    """
    meta class to allow interaction between :class:`datetime.datetime`
    objects create inside the :class:`hiro.Timeline` with those created
    outside it.
    """

    def __instancecheck__(cls, instance):
        return isinstance(instance, realdatetime)


class DateMeta(type):
    """
    meta class to allow interaction between :class:`datetime.date`
    objects create inside the :class:`hiro.Timeline` with those created
    outside it.
    """

    def __instancecheck__(cls, instance):
        return isinstance(instance, realdate)


class Datetime(realdatetime, metaclass=DatetimeMeta):
    """
    used to patch :class:`datetime.datetime` to follow the rules of the
    parent :class:`hiro.Timeline`
    """

    @classmethod
    def now(cls, tz=None):
        return cls.fromtimestamp(time.time(), tz)

    @classmethod
    def utcnow(cls):
        return cls.utcfromtimestamp(time.time())


class Date(realdate, metaclass=DateMeta):
    """
    used to patch :class:`datetime.date` to follow the rules of the
    parent :class:`hiro.Timeline`
    """

    __metaclass__ = DateMeta

    @classmethod
    def today(cls):
        return cls.fromtimestamp(time.time())



"""
setup.py for hiro

"""
__author__ = "Ali-Akber Saifee"
__email__ = "ali@indydevs.org.com"
__copyright__ = "Copyright 2023, Ali-Akber Saifee"

from setuptools import setup
import os

import versioneer

THIS_DIR = os.path.abspath(os.path.dirname(__file__))
REQUIREMENTS = [
    k for k in open(
        os.path.join(THIS_DIR, 'requirements', 'main.txt')
    ).read().splitlines() if k.strip()
]

extra = {}

setup(
    name='hiro',
    author=__author__,
    author_email=__email__,
    license="MIT",
    url="http://hiro.readthedocs.org",
    zip_safe=False,
    include_package_data=True,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=list(REQUIREMENTS),
    classifiers=[k for k in open('CLASSIFIERS').read().split('\n') if k],
    description='time manipulation utilities for testing in python',
    long_description=open('README.rst').read() + open('HISTORY.rst').read(),
    packages=["hiro"]
)



"""

"""
import time
from datetime import date, datetime


def sub_sample_2_1_now():
    return datetime.now()


def sub_sample_2_1_today():
    return date.today()


def sub_sample_2_1_sleep(seconds):
    time.sleep(seconds)


def sub_sample_2_1_time():
    return time.time()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('../../'))
sys.path.insert(0, os.path.abspath("./"))

import hiro
from theme_config import *

html_static_path = ["./_static"]
html_css_files = [
    "custom.css",
    "https://fonts.googleapis.com/css2?family=Fira+Code:wght@300;400;700&family=Fira+Sans:ital,wght@0,100;0,200;0,300;0,400;0,500;0,600;0,700;0,800;0,900;1,100;1,200;1,300;1,400;1,500;1,600;1,700;1,800;1,900&display=swap",
]

# -- Project information -----------------------------------------------------

if ".post0.dev" in hiro.__version__:
    version, ahead = hiro.__version__.split(".post0.dev")
else:
    version = hiro.__version__

project = 'hiro'
copyright = '2023, Ali-Akber Saifee'
author = 'Ali-Akber Saifee'
version = version
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    "sphinx_paramlinks",
    'sphinx.ext.viewcode'
]
# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

intersphinx_mapping = {
    "python": ("http://docs.python.org/", None)
}
